/***********************************************************
 * fft.c 
 ***********************************************************/

static void make_bitrev(int n, int bitrev[])
{
    int i = 0, j = 0, k;
    for (; ;) {
	bitrev[i] = j;
	if (++i >= n)  break;
	k = n/2;
	while (k <= j) { j -= k; k /= 2; }
	j += k;
    }
}

/*
 *  FFT (Cooley--Tukey's algorithm)
 *  n must be 2^p.
 *  x[k], y[k]  (k = 0, 1, 2, ... , |n|-1).
 *  n = 0 --> free sintbl, bitrev.
 *  n < 0 --> inverse Fourier tr.
 */
int fft(int n, double x[], double y[])
{
    const double PI = 3.14159265358979323846;
    static int    *bitrev = NULL;
    static int     last_n;
    int i, j, k, inverse;
    
    if (n < 0)  { n *= -1; inverse = 1; }
    else        { inverse = 0; }

    if (n != last_n || n == 0) {
        last_n = n;
	if (bitrev != NULL) free(bitrev);
	if (n == 0) return 0;
	bitrev = (int *) malloc(n*sizeof(int));
	if (bitrev == NULL) {
	    fprintf(stderr, "cannot allocate memory\n"); return 1;
	}
	make_bitrev(n, bitrev);
    }

    for (i = 0; i < n; i++) {
	j = bitrev[i];
	if (i < j) {
	    double tmp;
	    tmp = x[i]; x[i] = x[j]; x[j] = tmp;
	    tmp = y[i]; y[i] = y[j]; y[j] = tmp;
	}
    }
    for (k = 1; k < n; k = 2*k) {
	for (j = 0; j < k; j++) {
	    double 
		wx = cos(j*PI/k),
		wy = (inverse) ? -sin(j*PI/k) : sin(j*PI/k);
	    for (i = j; i < n; i += 2*k) {
		int ik = i + k;
		double dx =  wx*x[ik] + wy*y[ik];
		double dy = -wy*x[ik] + wx*y[ik];
		x[ik] = x[i] - dx;  x[i] += dx;
		y[ik] = y[i] - dy;  y[i] += dy;
	    }
	}
    }
    if (!inverse)
	for (i = 0; i < n; i++) { x[i] /= n; y[i] /= n;	}
    return 0;
}
