/*******************************************************************
 * radon.c: Symulation of CT_Scan of Head 
 *******************************************************************/

/*
 * Radon data: line integral of a function on R^{2}
 * rdn(w,p) = \int_{l} f(x,y) dl,  l: x*cos(w) + y*sin(w) = p
 *
 * Inversion Formula [Filtered Back Projection]:
 *
 * fbp(w,p) = \int rdn(w,q)*weight(p-q) dq
 *
 * f(x,y)   = \int_{[0,Pi]} fbp(w, x*cos(w)+y*sin(w)) dw
 *
 * weight(p)  = - (1/Pi) f.p.(1/p^2) = (1/2Pi) \int e^{ipr} |r| dr; [by Radon]
 * 
 * weight0(p) = - (1/Pi) (1 + cos(2Pi*p)) / (p^2 - (dp/2)^2);  [by Shepp Logan]
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "grlib.h"
#include "fft.c"

int width = 400, height = 400;

const double xmax = 3.0, ymax = 3.0;

#define IX(X)  ( .5*width *(1.0 + (X)/xmax) )
#define IY(Y)  ( .5*height*(1.0 - (Y)/ymax) )
#define X(IX)  (  ((IX) - (width /2))*xmax*2/width  )
#define Y(IY)  ( -((IY) - (height/2))*ymax*2/height )

#define PI 3.14159265358979323846

Color C = {255,255,255}, green = { 0,255, 0};

int    Nw, Np, Nl;
double dw, dp, dl;

/* Flag */
int have_rdn = 0;
int have_fbp = 0;
int show_integral = 0;
int use_fft = 0;

double **rdn,  **fbp;     /* Radon data, Filtered data */
double **abuf, **bbuf;    /* accumulation buf, back projection buf */
double *sn, *cn;          /* sin and cos table */

double head0(double x, double y);
double head1(double x, double y);
double head2(double x, double y);
double head3(double x, double y);
double head4(double x, double y);

double (*head)(double x, double y) = head0;

double weight(double p, double dp);

void initRadon(int N_w, int N_p);
void getRadon(void);
void getFbp0(void);
void getFbp1(void);
void reconstruct(void);
void convolution(double f1, double f2);
void accumulation(double **dat, double r1, double r2);
void drawHead(void);
void drawFilter(void);
void showIntegral(void);
void loop(void);

int main(int argc, char *argv[])
{
    InitCanvas(width, height, "Radon");

    initRadon(100, 256);
    drawHead();

    loop();
    return 0;
}

void initRadon(int N_w, int N_p)
{
    static int Nw_alloc = 0;
    int ix, iw;

    Nw = N_w;   Np = N_p;       Nl = width;
    dw = PI/Nw; dp = xmax*2/Np; dl = ymax*2/Nl;
    have_fbp = have_rdn = 0;

    if (sn) { free(cn); free(sn); }
    sn = (double *)malloc((Nw+1)*sizeof(double));
    cn = (double *)malloc((Nw+1)*sizeof(double));
    for (iw=0; iw<=Nw; iw++) {
	cn[iw] = cos(iw*dw); sn[iw] = sin(iw*dw);
    }

    if (rdn) {
	for (iw=0; iw<=Nw_alloc; iw++) { free(rdn[iw]); free(fbp[iw]); }
	free(rdn); free(fbp);
    }
    rdn = (double **)malloc((Nw+1)*sizeof(double *));
    fbp = (double **)malloc((Nw+1)*sizeof(double *));
    for (iw=0; iw<=Nw; iw++) {
	rdn[iw] = (double *)malloc((Np+1)*sizeof(double));
	fbp[iw] = (double *)malloc((Np+1)*sizeof(double));
    }
    Nw_alloc = Nw;

    if (abuf) {
	for (iw=0; iw<width; iw++) { free(abuf[iw]); free(bbuf[iw]); }
	free(abuf); free(bbuf);
    }
    abuf = (double **)malloc((width)*sizeof(double *));
    bbuf = (double **)malloc((width)*sizeof(double *));
    for (ix=0; ix<width; ix++) {
	abuf[ix] = (double *)malloc((height)*sizeof(double));
	bbuf[ix] = (double *)malloc((height)*sizeof(double));
    }
}

static void Pause(char *str, int pause)
{
    extern void mouse(int button, int state, double x, double y);

    SetColoredGray(&C, .0); FillRectangle(0, 0, width, 16);
    SetColoredGray(&C, .7); DrawString(10, 14, str);
    FlushBuffer();
    if (pause) {
	SetMouseFunc(NULL);
	while(!KbHit()); 
	SetMouseFunc(mouse);
    }
}

void getRadon(void)
{
    int iw, ip, il;

    SetDoubleBuffer(1);
    for (iw=0; iw<=Nw; iw++) {
	ClearCanvas();
	for (ip=0; ip<Np+1; ip++) {  
	    double sum = 0;
	    double xp = ip*dp - xmax, yp = -ymax;
	    double x0 = xp*cn[iw] - yp*sn[iw], y0 = xp*sn[iw] + yp*cn[iw];
	    double x1 = xp*cn[iw] + yp*sn[iw], y1 = xp*sn[iw] - yp*cn[iw];

	    for (il=Nl/20; il<Nl*19/20; il++) {
		double x = x0 + (x1-x0)*il/Nl, y = y0 + (y1-y0)*il/Nl;
		sum += head(x, y);

		if (show_integral && xp*xp<7) {
		    SetColoredGray(&C, .7*exp(-.7*sum*dl));
		    DrawPoint(IX(x), IY(y));
		}
	    }

	    rdn[iw][ip] = sum*dl;

	    if (xp*xp < 7) {
		SetColoredGray(&C, 0.8);
		DrawPoint(IX(x0*0.97+x1*0.03), IY(y0*0.97+y1*0.03));
		SetColoredGray(&C, 1.1 - exp(-.3*sum*dl));
		SetLineWidth((int)(0.52*dp*width/xmax + 1.5));
		DrawLine(IX(x0*0.08+x1*0.92), IY(y0*0.08+y1*0.92),
			 IX(x0*0.03+x1*0.97), IY(y0*0.03+y1*0.97));
	    } 
	}
	if (show_integral && iw==0) 
	    Pause("Hit Return to Continue.", 1);
	FlushBuffer();
    }
    have_rdn = 1;
    SetDoubleBuffer(0);
}

void showIntegral(void) { show_integral=1; getRadon(); show_integral=0; }

void getFbp0(void)
{
    int iw, ip, iq;

    if (!have_rdn) getRadon();
    Pause("Calculating FBP...", 0);
    for (iw=0; iw<=Nw; iw++) {
	for (ip=0; ip<=Np; ip++){
	    double sum = 0;
	    for (iq=0; iq<=Np; iq++) 
		sum = sum + rdn[iw][iq]*weight((ip-iq)*dp, dp);
	    fbp[iw][ip] = sum/(2*PI);
	}
	FillRectangle(125, 3, 125 + (width-125)*iw/Nw, 18);
	FlushBuffer();
    }
    have_fbp = 1;
}

void getFbp1(void)
{
    double *re = (double *)malloc(sizeof(double)*Np);
    double *im = (double *)malloc(sizeof(double)*Np);
    int iw, ip;
    
    if (!have_rdn) getRadon();
    Pause("Calculating FBP...", 0);
    for (iw=0; iw<=Nw; iw++) {
	for (ip=0; ip<Np; ip++) {
	    re[ip] = rdn[iw][(ip+Np/2)%Np];
	    im[ip] = 0;
	}

	fft(Np, re, im);
	for (ip=0; ip<Np; ip++){
	    double M = Np/PI/2/PI*sin(PI*ip/Np);
	    //double M = 1./2/PI*(Np/2 - fabs(ip-Np/2));
	    re[ip] *= M;
	    im[ip] *= M;
	}
	fft(-Np, re, im);

	for (ip=0; ip<Np; ip++)
	    fbp[iw][(ip+Np/2)%Np] = re[ip];

	FillRectangle(125, 3, 125 + (width-125)*iw/Nw, 18); FlushBuffer();
    }
    SetColoredGray(&C, 1);
    // FillRectangle(125, 3, 400, 18); FlushBuffer(); FlushBuffer();
    free(re); free(im);
    have_fbp = 1;
}

/**********************************************************************/

void reconstruct(void)
{
    double fn;
    int ix, iy, iw;

    if (use_fft)  getFbp1();
    else          getFbp0();
    ClearCanvas();
    for (iy=0; iy<height; iy++)
	for (ix=0; ix<width; ix++){
	    double sum = 0;
	    double x = X(ix), y = Y(iy);

	    if (x*x + y*y >= 9) continue;
	    for (iw=0; iw<Nw; iw++) {
		double p = (x*cn[iw] + y*sn[iw] + xmax)/dp;
		int   ip = (int) p;
		sum = sum + (ip+1-p)*fbp[iw][ip] + (p-ip)*fbp[iw][ip+1];
	    }
	    fn = sum*dw;
	    SetColoredGray(&C, fn);
	    DrawPoint(ix, iy);
	}
    FlushBuffer();
}

void convolution(double f1, double f2)
{
    Point2D rp[500], fp[500];
    int iw, ip;
    
    if (!have_fbp) getFbp0();
    for (iw=0; iw<=Nw; iw+=Nw/5){
	ClearCanvas();
	for (ip=0; ip<=Np; ip++){
	    double xp = ip*dp - xmax, yp = -ymax;
	    double x0 = xp*cn[iw] - yp*sn[iw], y0 = xp*sn[iw] + yp*cn[iw];
	    double x1 = xp*cn[iw] + yp*sn[iw], y1 = xp*sn[iw] - yp*cn[iw];
	    SetColoredGray(&C, .2*rdn[iw][ip]);
	    SetLineWidth((int)(0.25*dp*width/xmax + .5));
	    DrawLine(IX(x0), IY(y0), IX(x1), IY(y1));
	    rp[ip].x = IX(xp);
	    rp[ip].y = IY(f1*rdn[iw][ip]);
	}
	SetColoredGray(&C, .5); SetLineWidth(0);
	DrawLine(IX(-xmax), IY(0), IX(xmax), IY(0));
	SetColor(&green);
	DrawLines(rp, Np);
	FlushBuffer();
	Pause("Hit return.", 1);

	for (ip=0; ip<=Np; ip++){
	    double xp = ip*dp - xmax, yp = -ymax;
	    double x0 = xp*cn[iw] - yp*sn[iw], y0 = xp*sn[iw] + yp*cn[iw];
	    double x1 = xp*cn[iw] + yp*sn[iw], y1 = xp*sn[iw] - yp*cn[iw];
	    SetColoredGray(&C, .3*fbp[iw][ip]);
	    SetLineWidth((int)(0.25*dp*width/xmax + .5));
	    DrawLine(IX(x0), IY(y0), IX(x1), IY(y1));
	    fp[ip].x = IX(xp);
	    fp[ip].y = IY(f2*fbp[iw][ip]);
	}
	SetColoredGray(&C, .5);	SetLineWidth(0);
	DrawLine(IX(-xmax), IY(0), IX(xmax), IY(0));
	SetColoredGray(&C, .7);
	DrawLines(rp, Np);
	SetColor(&green); SetLineWidth(2);
	DrawLines(fp, Np);
	FlushBuffer();
	Pause("Hit return.", 1);
	if (iw == 0) SetDoubleBuffer(1);
    }
    Pause("END.", 0);
    SetDoubleBuffer(0);
}

void accumulation(double **dat, double r1, double r2)
//
#define forall(ix, iy)   for(iy=0;iy<height;iy++) for(ix=0;ix<width;ix++)
{
    int ix, iy, iw;
    
    if (!have_fbp) getFbp0();
    forall(ix,iy) { abuf[ix][iy] = 0; }
    for (iw=0; iw<Nw; iw++){
	forall(ix, iy) {
	    double x = X(ix), y = Y(iy);
	    double p = (x*cn[iw] + y*sn[iw] + xmax)/dp;
	    int   ip = (int) p;
	    if (ip>=0 && ip<Np-1) {		    
		bbuf[ix][iy] = (ip+1-p)*dat[iw][ip] + (p-ip)*dat[iw][ip+1];
		abuf[ix][iy]+= bbuf[ix][iy];
	    }
	}
	if (iw % (Nw/10) == 0 || iw == Nw-1) {
	    forall(ix, iy) {
		SetColoredGray(&C, r1*bbuf[ix][iy]);
		DrawPoint(ix, iy);
	    }
	    if (iw==0) Pause("Hit return to continue.", 1);
	    forall(ix, iy) {
		SetColoredGray(&C, r2*abuf[ix][iy]*dw*Nw*2/(iw+Nw));
		DrawPoint(ix, iy);
	    }
	}
    }
}

/**********************************************************************/

void drawFilter()
//
#define Weight0(x)  ( ((1-4*x*x) != 0) ? (4/PI)*(1+cos(2*PI*x))/(1-4*x*x) : 0 )
#define Weight1(x)  ( weight(x, 1.0) )
#define DrawAxes()  { DrawLine(IX(-xmax), IY(0), IX(xmax), IY(0)); \
		      DrawLine(IX(0), IY(-ymax), IX(0), IY(ymax)); }
{
    Point2D  w0[400], w1[400];
    int ix;

    for (ix=0; ix < width; ix++){
	double x = (ix - .5*width)/width;
	w0[ix].x = w1[ix].x = ix;
	w0[ix].y = IY(.8*Weight0(x*11));
	w1[ix].y = IY(.8*Weight1(x*11));
    }

    ClearCanvas();
    SetLineWidth(0);
    SetColoredGray(&C, .4); DrawAxes();
    SetColor(&green);       DrawLines(w0, width);
    Pause("Hit return",1);
    
    ClearCanvas();
    SetColoredGray(&C, .4); DrawAxes();
    SetColor(&green);       DrawLines(w1, width);
    Pause("Hit return",1);

    ClearCanvas();
    SetColoredGray(&C, .4); DrawAxes();
    SetColoredGray(&C, .6); DrawLines(w0, width);
    SetColor(&green); SetLineWidth(2); DrawLines(w1, width);

    FlushBuffer();
}

double weight(double x, double dx)	/* = Weight(x)*dx */
{
    double C = 4/PI/dx;
    double rx  = fabs(x/dx); 
    int    ix0 = (int)rx, ix1 = ix0+1;

    return C*((ix1-rx)/(1 - 4.*ix0*ix0) + (rx-ix0)/(1 - 4.*ix1*ix1));
}

/**********************************************************************/

void drawHead()
{
    int ix, iy;

    for (iy=0; iy<height; iy++)
	for (ix=0; ix<width ; ix++){
	    double x = xmax*(ix-0.5*width)*2.0/height;
	    double y = ymax*(.5*height-iy)*2.0/height;
	    SetColoredGray(&C, head(x, y));
	    DrawPoint(ix, iy);
	}
    FlushBuffer();
}

double head0(double x, double y)
//
#define Quad(x,y,A,B,C)  ((A)*(x)*(x) + (B)*(x)*(y) + (C)*(y)*(y))
{
    if (0.33*x*x + 0.2*y*y > 1.00)			return 0.0;
    if (0.33*x*x + 0.2*y*y > 0.92)			return 1.0;
    if (2*x*x + 1.25*(y-1.0)*(y-1.0) < 0.70)  		return 0.9;
    if (x*x + (y+0.1)*(y+0.1) < 0.02)      		return 1.0;
    if (Quad(x+0.8, y-0.2, 3.5, 1.5, 1.0) < 0.70)	return 0.4;
    if (Quad(x-0.8, y-0.2, 2.5,-1.0, 1.0) < 0.40)	return 0.4;
    if (Quad(x-1.2, y+1.2, 3.0,-1.0, 1.0) < 0.20)	return 0.8;
    else 	                                        return 0.7;
}

double head1(double x, double y)
{
    return ((x)*(x) + (y-.1)*(y-.1) < 0.9) ? .8 : 0;
}

double head2(double x, double y)
{
    if ((x)*(x) + (y-.1)*(y-.1) < 0.6)  return .4;
    if ((x)*(x) + (y-.1)*(y-.1) < 0.9)  return .8;
    else                                return .0;
}

double head3(double x, double y)
{
    return (fabs(x) < 1.2 && fabs(y) < 1.2) ? .8 : 0;
}

double head4(double x, double y)
{
    if (fabs(x) < 1.0 && fabs(y) < 1.0)	return .2;
    if (fabs(x) < 1.2 && fabs(y) < 1.2)	return .9;
    else				return .0;
}

/**********************************************************************/

void actionfunc(int c)
{
    static char str[100];
    static char fname[100];
    show_integral = 0;

    switch (c){
    case 'q': CloseCanvas();  exit(0);

    case 'h': drawHead();                return;
    case 'r': getRadon();                return;
    case 'R': reconstruct();             return;
    case 'f': drawFilter();              return;
    case 'i': showIntegral();            return;
    case 'c': convolution(.5, .5);       return;
    case 'a': accumulation(fbp, .3, 1.); return;
    case 'o': accumulation(rdn, .1, .1); return;
    case 'F': use_fft = !use_fft; have_fbp = 0; return;
    case 's': sprintf(fname, "%dx%d.bmp", Nw, Np); SaveImageAs(fname); return;

    /* change params */
    #define DispParams() { sprintf(str,"Nw=%3d,Np=%3d",Nw,Np); Pause(str,0); }
    case 1001: Nw =  30; Np =  64; initRadon(Nw, Np); DispParams(); return;
    case 1002: Nw =  60; Np = 128; initRadon(Nw, Np); DispParams(); return;
    case 1003: Nw = 100; Np = 256; initRadon(Nw, Np); DispParams(); return;
    case 1004: Nw = 200; Np = 512; initRadon(Nw, Np); DispParams(); return;

    /* change object */
    case 2000: head = head0; have_fbp = have_rdn = 0; drawHead(); return;
    case 2001: head = head1; have_fbp = have_rdn = 0; drawHead(); return;
    case 2002: head = head2; have_fbp = have_rdn = 0; drawHead(); return;
    case 2003: head = head3; have_fbp = have_rdn = 0; drawHead(); return;
    case 2004: head = head4; have_fbp = have_rdn = 0; drawHead(); return;

    default:
	return;
    }
}

MenuItem
params[]={
    {"Nw= 30, Np= 50", 1001},
    {"Nw= 60, Np=128", 1002},
    {"Nw=100, Np=256", 1003},
    {"Nw=200, Np=512", 1004},
    { NULL, 0}
};
Menu chpara={params};

MenuItem
objs[]={
    {"Shepp-Logan's Phantom", 2000},
    {"Disk0", 2001},
    {"Disk1", 2002},
    {"Box0" , 2003},
    {"Box1" , 2004},
    { NULL, 0}
};
Menu chobj={objs};

MenuItem
demos[]={
    {"Line integral",   'i'},
    {"Convolution",     'c'},
    {"FBP accumulation", 'a'},
    {"Old tomography",  'o'},
    { NULL, 0}
};
Menu demomenu={demos};

MenuItem
topitems[]={
    {"Draw Original",         'h'},
    {"Calc Radon",            'r'},
    {"Reconstruction by FBP", 'R'},
    {"Demos",               0, &demomenu},
    {"Change Parameter",    0, &chpara},
    {"Change Object",       0, &chobj},
    {"Draw Filter",           'f'},
    {"Use FFT (toggle)",      'F'},
    {"Save Image as \'(Nw)x(Np).bmp\'", 's'},
    {"Quit", 'q'},
    { NULL ,  0 }
};
Menu topmenu={topitems};

void mouse(int button, int state, double x, double y)
{
    if (state == MOUSE_DOWN) { PopupMenu(&topmenu, actionfunc); }
}

void loop(void)
{
    SetMouseFunc(mouse);
    for (; ;)
	if (KbHit()) { actionfunc(GetKey()); }
}
